import vigra

import vigra.graphs as vgraph

import graph as agraph

import numpy
import scipy.ndimage
from MCSolverImpl import *
from Tools import *
from sklearn.ensemble import RandomForestClassifier



def hessianEv(img, sigma):
    if img.squeeze().ndim  == 2:
        return vigra.filters.hessianOfGaussianEigenvalues(img, sigma)[:,:,0]
    else:
        out = numpy.zeros(img.shape, dtype='float32')
        for z in range(img.shape[2]):
            out[:, :, z] = vigra.filters.hessianOfGaussianEigenvalues(img[:,:,z], sigma)[:,:,0]
        return out



def gaussianSmooth(img, sigma):
    if img.squeeze().ndim  == 2:
        return vigra.filters.gaussianSmoothing(img, sigma)
    else:
        out = numpy.zeros(img.shape, dtype='float32')
        for z in range(img.shape[2]):
            out[:, :, z] = vigra.filters.gaussianSmoothing(img[:,:,z], sigma)
        return out




@cacher_hdf5(ignoreNumpyArrays=True)
def clusteringFeatures(ds, segId, extraUV, edgeIndicator, liftedNeighborhood):

    print "Computing clustering features for lifted neighborhood", liftedNeighborhood

    originalGraph =  ds._rag(segId)
    extraUV = numpy.require(extraUV,dtype='uint32')
    uvOriginal = originalGraph.uvIds()
    liftedGraph =  vgraph.listGraph(originalGraph.nodeNum)
    liftedGraph.addEdges(uvOriginal)
    liftedGraph.addEdges(extraUV)

    uvLifted = liftedGraph.uvIds()
    foundEdges = originalGraph.findEdges(uvLifted)
    foundEdges[foundEdges>=0] = 0
    foundEdges *= -1

    nAdditionalEdges = liftedGraph.edgeNum - originalGraph.edgeNum
    whereLifted = numpy.where(foundEdges==1)[0].astype('uint32')
    assert len(whereLifted) == nAdditionalEdges
    assert foundEdges.sum() == nAdditionalEdges


    allFeat = []
    eLen = vgraph.getEdgeLengths(originalGraph)
    nodeSizes_ = vgraph.getNodeSizes(originalGraph)
    #for wardness in [0.01, 0.6, 0.7]:
    for wardness in [0.01, 0.1, 0.2, 0.3 ,0.4, 0.5, 0.6, 0.7]:

        edgeLengthsNew = numpy.concatenate([eLen,numpy.zeros(nAdditionalEdges)]).astype('float32')
        edgeIndicatorNew = numpy.concatenate([edgeIndicator,numpy.zeros(nAdditionalEdges)]).astype('float32')

        nodeSizes = nodeSizes_.copy()
        nodeLabels = vgraph.graphMap(originalGraph,'node',dtype='uint32')

        nodeFeatures = vgraph.graphMap(liftedGraph,'node',addChannelDim=True)
        nodeFeatures[:]=0


        outWeight=vgraph.graphMap(liftedGraph,item='edge',dtype=numpy.float32)


        mg = vgraph.mergeGraph(liftedGraph)
        clusterOp = vgraph.minEdgeWeightNodeDist(mg,
            edgeWeights=edgeIndicatorNew,
            edgeLengths=edgeLengthsNew,
            nodeFeatures=nodeFeatures,
            nodeSizes=nodeSizes,
            nodeLabels=nodeLabels,
            beta=0.5,
            metric='l1',
            wardness=wardness,
            outWeight=outWeight
        )


        clusterOp.setLiftedEdges(whereLifted)


        hc = vgraph.hierarchicalClustering(clusterOp,nodeNumStopCond=1,
                                            buildMergeTreeEncoding=False)
        hc.cluster()

        assert mg.nodeNum == 1
        assert mg.edgeNum == 0
        tweight = edgeIndicatorNew.copy()
        hc.ucmTransform(tweight)

        whereInLifted = liftedGraph.findEdges(extraUV)
        assert whereInLifted.min() >= 0
        feat = tweight[whereInLifted]
        assert feat.shape[0] == extraUV.shape[0]
        allFeat.append(feat[:,None])

    weights = numpy.concatenate(allFeat,axis=1)
    print weights.shape
    mean = numpy.mean(weights,axis=1)[:,None]
    stddev = numpy.std(weights,axis=1)[:,None]
    allFeat = numpy.concatenate([weights,mean,stddev],axis=1)

    return numpy.nan_to_num(allFeat)





@cacher_hdf5(ignoreNumpyArrays=True)
def compute_lifted_feature_multicut(ds, segId,
        pLocal, pipelineParam, uvIds, liftedNeighborhood):

    print "Computing multcut features for lifted neighborhood", liftedNeighborhood

    from concurrent import futures

    # variables for the multicuts
    uv_ids     = ds._adjacent_segments(segId)
    seg_id_max = ds.seg(segId).max()
    n_var = seg_id_max + 1

    # scaling for energies
    edge_probs = pLocal.copy()

    edge_indications = ds.edge_indications(segId)
    edge_areas       = ds._rag(segId).edgeLengths()

    def single_mc(pLocal, edge_indications, edge_areas,
            uv_ids, n_var, seg_id, pipelineParam, beta, weight):

        # copy the probabilities
        probs = pLocal.copy()
        p_min = 0.001
        p_max = 1. - p_min
        probs = (p_max - p_min) * edge_probs + p_min
        # probs to energies
        energies = np.log( (1. - probs) / probs ) + np.log( (1. - beta) / beta )

        # weight the energies
        if pipelineParam.weighting_scheme == "z":
            print "Weighting Z edges"
            # z - edges are indicated with 0 !
            area_z_max = float( np.max( edge_areas[edge_indications == 0] ) )
            # we only weight the z edges !
            w = weight * edge_areas[edge_indications == 0] / area_z_max
            energies[edge_indications == 0] = np.multiply(w, energies[edge_indications == 0])

        elif pipelineParam.weighting_scheme == "xyz":
            print "Weighting xyz edges"
            # z - edges are indicated with 0 !
            area_z_max = float( np.max( edge_areas[edge_indications == 0] ) )
            len_xy_max = float( np.max( edge_areas[edge_indications == 1] ) )
            # weight the z edges !
            w_z = weight * edge_areas[edge_indications == 0] / area_z_max
            energies[edge_indications == 0] = np.multiply(w_z, energies[edge_indications == 0])
            # weight xy edges
            w_xy = weight * edge_areas[edge_indications == 1] / len_xy_max
            energies[edge_indications == 1] = np.multiply(w_xy, energies[edge_indications == 1])

        elif pipelineParam.weighting_scheme == "all":
            print "Weighting all edges"
            area_max = float( np.max( edge_areas ) )
            w = weight * edge_areas / area_max
            energies = np.multiply(w, energies)

        # get the energies (need to copy code here, because we can't use caching in threads)
        mc_node, mc_edges, mc_energy, t_inf = multicut_fusionmoves(
                n_var, uv_ids,
                energies, pipelineParam)

        return mc_node

    #single_mc( pLocal, edge_indications, edge_areas,
    #    uv_ids, n_var, segId, pipelineParam, 0.5, 16 )

    with futures.ThreadPoolExecutor(max_workers=pipelineParam.n_threads) as executor:
    #with futures.ThreadPoolExecutor(max_workers=2) as executor:
        tasks = []
        for beta in (0.4, 0.45, 0.5, 0.55, 0.65):
            for w in (12, 16, 25):
                tasks.append( executor.submit( single_mc, pLocal, edge_indications, edge_areas,
                    uv_ids, n_var, segId, pipelineParam, beta, w ) )

    mc_nodes = [future.result() for future in tasks]

    # map multicut result to lifted edges
    allFeat = [ ( mc_node[uvIds[:, 0]] !=  mc_node[uvIds[:, 1]] )[:,None] for mc_node in mc_nodes]

    mcStates = numpy.concatenate(allFeat, axis=1)
    stateSum = numpy.sum(mcStates,axis=1)
    return numpy.concatenate([mcStates,stateSum[:,None]],axis=1)


def lifted_feature_aggregator(ds, featureList,
        pLocal, pipelineParam, uvIds, segId ):

    assert len(featureList) > 0
    for feat in featureList:
        assert feat in ("mc", "cluster","reg"), feat

    features = []
    if "mc" in featureList:
        features.append( compute_lifted_feature_multicut(ds, segId,
            pLocal, pipelineParam, uvIds, pipelineParam.lifted_neighborhood) )
    if "cluster" in featureList:
        features.append( clusteringFeatures(ds, segId,
            uvIds, pLocal, pipelineParam.lifted_neighborhood ) )
    if "reg" in featureList:
        features.append( ds.region_features(segId, 0, uvIds, pipelineParam.lifted_neighborhood) )

    return numpy.concatenate( features, axis=1 )



@cacher_hdf5()
def compute_and_save_lifted_nh(ds, segId, liftedNeighborhood):

    rag = ds._rag(segId)
    originalGraph = agraph.Graph(rag.nodeNum)
    originalGraph.insertEdges(rag.uvIds())

    print ds.ds_name
    print "Computing lifted neighbors for range:", liftedNeighborhood
    lm= agraph.liftedMcModel(originalGraph)
    agraph.addLongRangeNH(lm , liftedNeighborhood)
    uvIds = lm.liftedGraph().uvIds()

    return uvIds[rag.edgeNum:,:]


@cacher_hdf5()
def node_z_coord(ds, segId):

    rag = ds._rag(segId)
    labels = rag.labels
    labels = labels.squeeze()

    nz = numpy.zeros(rag.maxNodeId +1, dtype='uint32')

    for z in range(labels.shape[2]):
        lz = labels[:,:,z]
        nz[lz] = z

    return nz


@cacher_hdf5(ignoreNumpyArrays=True)
def lifted_fuzzy_gt(ds, segId, uvIds):

    gt = ds.gt()
    oseg = ds.seg(segId)


    fuzzyLiftedGt = agraph.candidateSegToRagSeg(
    oseg.astype('uint32'), gt.astype('uint32'),
        uvIds.astype(np.uint64))

    return fuzzyLiftedGt

#@cacher_hdf5(ignoreNumpyArrays=True)
def lifted_hard_gt(ds, segId, uvIds):

    rag = ds._rag(segId)
    gt = ds.gt()
    nodeGt,_ =  rag.projectBaseGraphGt(gt)

    return nodeGt[uvIds[:,0]] != nodeGt[uvIds[:,1]],nodeGt


# TODO we should cache this for rerunning experiments
# (will take very long for larger ds like snemi)
#@cacher_hdf5(ignoreNumpyArrays=True)
def doActualTrainingAndPrediction(dsTrain, X, Y, F, pipelineParam):
    rf = RandomForestClassifier(n_estimators = pipelineParam.n_trees,
            n_jobs=pipelineParam.n_threads,  oob_score=True)
    rf.fit(X, Y.astype('uint32'))
    print "Trained RF on lifted edges:"
    print "OOB-Error:", 1. - rf.oob_score_

    pTest = rf.predict_proba(F)

    return pTest


def learn_and_predict_lifted(dsTrain, dsTest,
        segIdTrain, segIdTest,
        feature_list_lifted,
        pLocalTrain, pLocalTest,
        pipelineParam ):


    uvIdsTrain = compute_and_save_lifted_nh(dsTrain, segIdTrain, pipelineParam.lifted_neighborhood)
    uvIdsTest = compute_and_save_lifted_nh(dsTest, segIdTest, pipelineParam.lifted_neighborhood)


    nzTrain = node_z_coord(dsTrain, segIdTrain)
    nzTest  = node_z_coord(dsTest, segIdTest)


    # compute the features for the training set
    fTrain = lifted_feature_aggregator(dsTrain, feature_list_lifted,
            pLocalTrain, pipelineParam, uvIdsTrain, segIdTrain)
    fTest = lifted_feature_aggregator(dsTest, feature_list_lifted,
            pLocalTest, pipelineParam, uvIdsTest, segIdTest)

    fuzzyLiftedGt = lifted_fuzzy_gt(dsTrain, segIdTrain, uvIdsTrain)
    hardLiftedGt,nodeGt = lifted_hard_gt(dsTrain, segIdTrain, uvIdsTrain)


    # check which ov the edges is in plane
    zU = nzTrain[uvIdsTrain[:,0]]
    zV = nzTrain[uvIdsTrain[:,1]]


    #where in plane
    if pipelineParam.learn_2d:
        whereGtMaskA = (zU == zV)
        whereGtMaskB = (nodeGt[uvIdsTrain[:,0]] != 0)
        whereGtMaskC = (nodeGt[uvIdsTrain[:,1]] != 0)
        whereGtMask = whereGtMaskA | whereGtMaskB
        whereGtMask = whereGtMask | whereGtMaskC
        #whereGtMask = (zU != 0 and  zV !=0 and whereGtMaskA==True)
    else:
        whereGtMaskA = (nodeGt[uvIdsTrain[:,0]] != 0)
        whereGtMaskB = (nodeGt[uvIdsTrain[:,1]] != 0)
        whereGtMask = whereGtMaskA | whereGtMaskB
    whereGt = numpy.where(whereGtMask==True)[0]


    YH = hardLiftedGt[whereGt]
    YF = fuzzyLiftedGt[whereGt]
    X =  fTrain[whereGt,:]


    Y = numpy.zeros(YF.shape)
    Y[YF>0.5] = 1

    pTest = doActualTrainingAndPrediction(dsTrain, X,YH, fTest, pipelineParam)

    return pTest[:,1], uvIdsTest,nzTest




def optimizeLifted(dsTest, model, rag, starting_point = None):
    print "Optimizing lifted model"

    # if no starting point is given, start with ehc solver
    if starting_point is None:
        # settings ehc
        print "Starting from ehc solver result"
        settingsGa = agraph.settingsLiftedGreedyAdditive(model)
        ga = agraph.liftedGreedyAdditive(model, settingsGa)
        ws = ga.run()
    # else, we use the starting point that is given as argument
    else:
        print "Starting from external starting point"
        ws = starting_point.astype('uint8')

    # setttings for kl
    settingsKl = agraph.settingsLiftedKernighanLin(model)
    kl = agraph.liftedKernighanLin(model, settingsKl)
    ws2 = kl.run(ws)


    # FM RAND
    # settings for proposal generator
    settingsProposalGen = agraph.settingsProposalsFusionMovesRandomizedProposals(model)
    settingsProposalGen.sigma = 10.5
    settingsProposalGen.nodeLimit = 0.5

    # settings for solver itself
    settings = agraph.settingsFusionMoves(settingsProposalGen)
    settings.maxNumberOfIterations = 1
    settings.nParallelProposals = 2
    settings.reduceIterations = 1
    settings.seed = 42
    settings.verbose = 0
    # solver
    solver = agraph.fusionMoves(model, settings)
    ws3 = solver.run(ws2)


    # FM SG
    # settings for proposal generator
    settingsProposalGen = agraph.settingsProposalsFusionMovesSubgraphProposals(model)
    settingsProposalGen.subgraphRadius = 5

    # settings for solver itself
    settings = agraph.settingsFusionMoves(settingsProposalGen)
    settings.maxNumberOfIterations = 3
    settings.nParallelProposals = 10
    settings.reduceIterations = 0
    settings.seed = 43
    settings.verbose = 0
    # solver
    solver = agraph.fusionMoves(model, settings)
    out = solver.run(ws3)

    nodeLabels = model.edgeLabelsToNodeLabels(out)
    nodeLabels = nodeLabels.astype('uint32')

    return nodeLabels
